@@ -1,12 +1,17 @@
-/* Parser interface for DOM-based parser (libxml) rather than
-   stream-based SAX-type parser */
-
+/*
+ * $PostgreSQL: pgsql/contrib/xml2/xpath.c,v 1.23.2.4 2010/03/03 19:10:29 tgl Exp $
+ *
+ * Parser interface for DOM-based parser (libxml) rather than
+ * stream-based SAX-type parser
+ */
 #include "postgres.h"
-#include "fmgr.h"
+
 #include "executor/spi.h"
+#include "fmgr.h"
 #include "funcapi.h"
-#include "miscadmin.h"
 #include "lib/stringinfo.h"
+#include "miscadmin.h"
+#include "utils/builtins.h"
 #include "utils/xml.h"
 
 /* libxml includes */
@@ -35,15 +40,6 @@ Datum		xpath_table(PG_FUNCTION_ARGS);
 
 void		pgxml_parser_init(void);
 
-/* workspace for pgxml_xpath() */
-
-typedef struct
-{
-	xmlDocPtr	doctree;
-	xmlXPathContextPtr ctxt;
-	xmlXPathObjectPtr res;
-} xpath_workspace;
-
 /* local declarations */
 
 static xmlChar *pgxmlNodeSetToText(xmlNodeSetPtr nodeset,
@@ -55,12 +51,7 @@ static text *pgxml_result_to_text(xmlXPathObjectPtr res, xmlChar *toptag,
 
 static xmlChar *pgxml_texttoxmlchar(text *textstring);
 
-static xmlXPathObjectPtr pgxml_xpath(text *document, xmlChar *xpath,
-									 xpath_workspace *workspace);
-
-static void cleanup_workspace(xpath_workspace *workspace);
-
-#define GET_STR(textp) DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(textp)))
+static xmlXPathObjectPtr pgxml_xpath(text *document, xmlChar *xpath);
 
 
 /*
@@ -110,7 +101,6 @@ xml_encode_special_chars(PG_FUNCTION_ARGS)
 {
 	text	   *tin = PG_GETARG_TEXT_P(0);
 	text	   *tout;
-	int32		ressize;
 	xmlChar    *ts,
 			   *tt;
 
@@ -120,10 +110,7 @@ xml_encode_special_chars(PG_FUNCTION_ARGS)
 
 	pfree(ts);
 
-	ressize = strlen((char *) tt);
-	tout = (text *) palloc(ressize + VARHDRSZ);
-	memcpy(VARDATA(tout), tt, ressize);
-	SET_VARSIZE(tout, ressize + VARHDRSZ);
+	tout = cstring_to_text((char *) tt);
 
 	xmlFree(tt);
 
@@ -213,14 +200,7 @@ pgxmlNodeSetToText(xmlNodeSetPtr nodeset,
 static xmlChar *
 pgxml_texttoxmlchar(text *textstring)
 {
-	xmlChar    *res;
-	int32		txsize;
-
-	txsize = VARSIZE(textstring) - VARHDRSZ;
-	res = (xmlChar *) palloc(txsize + 1);
-	memcpy((char *) res, VARDATA(textstring), txsize);
-	res[txsize] = '\0';
-	return res;
+	return (xmlChar *) text_to_cstring(textstring);
 }
 
 /* Publicly visible XPath functions */
@@ -234,22 +214,25 @@ PG_FUNCTION_INFO_V1(xpath_nodeset);
 Datum
 xpath_nodeset(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
-	xmlChar    *toptag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
-	xmlChar    *septag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(3));
-	xmlChar    *xpath;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	xmlChar    *xpath,
+			   *toptag,
+			   *septag;
+	int32		pathsize;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	toptag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
+	septag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(3));
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, toptag, septag, NULL);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 toptag, septag, NULL);
 
 	pfree(xpath);
 
@@ -267,21 +250,23 @@ PG_FUNCTION_INFO_V1(xpath_list);
 Datum
 xpath_list(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
-	xmlChar    *plainsep = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
-	xmlChar    *xpath;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	xmlChar    *xpath,
+			   *plainsep;
+	int32		pathsize;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	plainsep = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, NULL, NULL, plainsep);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 NULL, NULL, plainsep);
 
 	pfree(xpath);
 
@@ -296,13 +281,13 @@ PG_FUNCTION_INFO_V1(xpath_string);
 Datum
 xpath_string(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
 	int32		pathsize;
-	text	   *xpres;
-	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+	text	   *xpathsupp,
+			   *xpres;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 
 	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
@@ -313,16 +298,13 @@ xpath_string(PG_FUNCTION_ARGS)
 	/* We could try casting to string using the libxml function? */
 
 	xpath = (xmlChar *) palloc(pathsize + 9);
-	strncpy((char *) xpath, "string(", 7);
 	memcpy((char *) (xpath + 7), VARDATA(xpathsupp), pathsize);
+	strncpy((char *) xpath, "string(", 7);
 	xpath[pathsize + 7] = ')';
 	xpath[pathsize + 8] = '\0';
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
-	xpres = pgxml_result_to_text(res, NULL, NULL, NULL);
-
-	cleanup_workspace(&workspace);
+	xpres = pgxml_result_to_text(pgxml_xpath(PG_GETARG_TEXT_P(0), xpath),
+								 NULL, NULL, NULL);
 
 	pfree(xpath);
 
@@ -337,17 +319,21 @@ PG_FUNCTION_INFO_V1(xpath_number);
 Datum
 xpath_number(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
+	int32		pathsize;
+	text	   *xpathsupp;
 	float4		fRes;
+
 	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
+	res = pgxml_xpath(PG_GETARG_TEXT_P(0), xpath);
 	pfree(xpath);
 
 	if (res == NULL)
@@ -355,8 +341,6 @@ xpath_number(PG_FUNCTION_ARGS)
 
 	fRes = xmlXPathCastToNumber(res);
 
-	cleanup_workspace(&workspace);
-
 	if (xmlXPathIsNaN(fRes))
 		PG_RETURN_NULL();
 
@@ -369,17 +353,21 @@ PG_FUNCTION_INFO_V1(xpath_bool);
 Datum
 xpath_bool(PG_FUNCTION_ARGS)
 {
-	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
 	xmlChar    *xpath;
+	int32		pathsize;
+	text	   *xpathsupp;
 	int			bRes;
+
 	xmlXPathObjectPtr res;
-	xpath_workspace workspace;
+
+	/* PG_GETARG_TEXT_P(0) is document buffer */
+	xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+
+	pathsize = VARSIZE(xpathsupp) - VARHDRSZ;
 
 	xpath = pgxml_texttoxmlchar(xpathsupp);
 
-	res = pgxml_xpath(document, xpath, &workspace);
-
+	res = pgxml_xpath(PG_GETARG_TEXT_P(0), xpath);
 	pfree(xpath);
 
 	if (res == NULL)
@@ -387,8 +375,6 @@ xpath_bool(PG_FUNCTION_ARGS)
 
 	bRes = xmlXPathCastToBoolean(res);
 
-	cleanup_workspace(&workspace);
-
 	PG_RETURN_BOOL(bRes);
 }
 
@@ -397,61 +383,49 @@ xpath_bool(PG_FUNCTION_ARGS)
 /* Core function to evaluate XPath query */
 
 static xmlXPathObjectPtr
-pgxml_xpath(text *document, xmlChar *xpath, xpath_workspace *workspace)
+pgxml_xpath(text *document, xmlChar *xpath)
 {
-	int32		docsize = VARSIZE(document) - VARHDRSZ;
+	xmlDocPtr	doctree;
+	xmlXPathContextPtr ctxt;
 	xmlXPathObjectPtr res;
 	xmlXPathCompExprPtr comppath;
+	int32		docsize;
 
-	workspace->doctree = NULL;
-	workspace->ctxt = NULL;
-	workspace->res = NULL;
+	docsize = VARSIZE(document) - VARHDRSZ;
 
 	pgxml_parser_init();
 
-	workspace->doctree = xmlParseMemory((char *) VARDATA(document), docsize);
-	if (workspace->doctree == NULL)
+	doctree = xmlParseMemory((char *) VARDATA(document), docsize);
+	if (doctree == NULL)
 		return NULL;			/* not well-formed */
 
-	workspace->ctxt = xmlXPathNewContext(workspace->doctree);
-	workspace->ctxt->node = xmlDocGetRootElement(workspace->doctree);
+	ctxt = xmlXPathNewContext(doctree);
+	ctxt->node = xmlDocGetRootElement(doctree);
 
 	/* compile the path */
 	comppath = xmlXPathCompile(xpath);
 	if (comppath == NULL)
 	{
-		cleanup_workspace(workspace);
+		xmlFreeDoc(doctree);
 		xml_ereport(ERROR, ERRCODE_EXTERNAL_ROUTINE_EXCEPTION,
 					"XPath Syntax Error");
 	}
 
 	/* Now evaluate the path expression. */
-	res = xmlXPathCompiledEval(comppath, workspace->ctxt);
-	workspace->res = res;
-
+	res = xmlXPathCompiledEval(comppath, ctxt);
 	xmlXPathFreeCompExpr(comppath);
 
 	if (res == NULL)
-		cleanup_workspace(workspace);
+	{
+		xmlXPathFreeContext(ctxt);
+		xmlFreeDoc(doctree);
 
+		return NULL;
+	}
+	/* xmlFreeDoc(doctree); */
 	return res;
 }
 
-/* Clean up after processing the result of pgxml_xpath() */
-static void
-cleanup_workspace(xpath_workspace *workspace)
-{
-	if (workspace->res)
-		xmlXPathFreeObject(workspace->res);
-	workspace->res = NULL;
-	if (workspace->ctxt)
-		xmlXPathFreeContext(workspace->ctxt);
-	workspace->ctxt = NULL;
-	if (workspace->doctree)
-		xmlFreeDoc(workspace->doctree);
-	workspace->doctree = NULL;
-}
-
 static text *
 pgxml_result_to_text(xmlXPathObjectPtr res,
 					 xmlChar *toptag,
@@ -459,7 +433,6 @@ pgxml_result_to_text(xmlXPathObjectPtr res,
 					 xmlChar *plainsep)
 {
 	xmlChar    *xpresstr;
-	int32		ressize;
 	text	   *xpres;
 
 	if (res == NULL)
@@ -483,10 +456,7 @@ pgxml_result_to_text(xmlXPathObjectPtr res,
 	}
 
 	/* Now convert this result back to text */
-	ressize = strlen((char *) xpresstr);
-	xpres = (text *) palloc(ressize + VARHDRSZ);
-	memcpy(VARDATA(xpres), xpresstr, ressize);
-	SET_VARSIZE(xpres, ressize + VARHDRSZ);
+	xpres = cstring_to_text((char *) xpresstr);
 
 	/* Free various storage */
 	xmlFree(xpresstr);
@@ -504,11 +474,11 @@ Datum
 xpath_table(PG_FUNCTION_ARGS)
 {
 	/* Function parameters */
-	char	   *pkeyfield = GET_STR(PG_GETARG_TEXT_P(0));
-	char	   *xmlfield = GET_STR(PG_GETARG_TEXT_P(1));
-	char	   *relname = GET_STR(PG_GETARG_TEXT_P(2));
-	char	   *xpathset = GET_STR(PG_GETARG_TEXT_P(3));
-	char	   *condition = GET_STR(PG_GETARG_TEXT_P(4));
+	char	   *pkeyfield = text_to_cstring(PG_GETARG_TEXT_PP(0));
+	char	   *xmlfield = text_to_cstring(PG_GETARG_TEXT_PP(1));
+	char	   *relname = text_to_cstring(PG_GETARG_TEXT_PP(2));
+	char	   *xpathset = text_to_cstring(PG_GETARG_TEXT_PP(3));
+	char	   *condition = text_to_cstring(PG_GETARG_TEXT_PP(4));
 
 	/* SPI (input tuple) support */
 	SPITupleTable *tuptable;
@@ -571,7 +541,9 @@ xpath_table(PG_FUNCTION_ARGS)
 	 * Create the tuplestore - work_mem is the max in-memory size before a
 	 * file is created on disk to hold it.
 	 */
-	tupstore = tuplestore_begin_heap(true, false, work_mem);
+	tupstore =
+		tuplestore_begin_heap(rsinfo->allowedModes & SFRM_Materialize_Random,
+							  false, work_mem);
 
 	MemoryContextSwitchTo(oldcontext);
 
@@ -694,7 +666,7 @@ xpath_table(PG_FUNCTION_ARGS)
 		/* Parse the document */
 		if (xmldoc)
 			doctree = xmlParseMemory(xmldoc, strlen(xmldoc));
-		else					/* treat NULL as not well-formed */
+		else	/* treat NULL as not well-formed */
 			doctree = NULL;
 
 		if (doctree == NULL)
