@@ -3,12 +3,12 @@
  * pg_shdepend.c
  *	  routines to support manipulation of the pg_shdepend relation
  *
- * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/catalog/pg_shdepend.c,v 1.23.2.3 2010/07/03 13:53:38 rhaas Exp $
+ *	  $PostgreSQL: pgsql/src/backend/catalog/pg_shdepend.c,v 1.34 2009/06/11 14:48:55 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -26,8 +26,6 @@
 #include "catalog/pg_language.h"
 #include "catalog/pg_namespace.h"
 #include "catalog/pg_operator.h"
-#include "catalog/pg_opclass.h"
-#include "catalog/pg_opfamily.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_shdepend.h"
 #include "catalog/pg_tablespace.h"
@@ -38,10 +36,12 @@
 #include "commands/schemacmds.h"
 #include "commands/tablecmds.h"
 #include "commands/typecmds.h"
+#include "storage/lmgr.h"
 #include "miscadmin.h"
 #include "utils/acl.h"
 #include "utils/fmgroids.h"
 #include "utils/syscache.h"
+#include "utils/tqual.h"
 
 
 typedef enum
@@ -55,13 +55,17 @@ static int getOidListDiff(Oid *list1, int nlist1, Oid *list2, int nlist2,
 			   Oid **diff);
 static Oid	classIdGetDbId(Oid classId);
 static void shdepLockAndCheckObject(Oid classId, Oid objectId);
-static void shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
+static void shdepChangeDep(Relation sdepRel,
+			   Oid classid, Oid objid, int32 objsubid,
 			   Oid refclassid, Oid refobjid,
 			   SharedDependencyType deptype);
-static void shdepAddDependency(Relation sdepRel, Oid classId, Oid objectId,
+static void shdepAddDependency(Relation sdepRel,
+				   Oid classId, Oid objectId, int32 objsubId,
 				   Oid refclassId, Oid refobjId,
 				   SharedDependencyType deptype);
-static void shdepDropDependency(Relation sdepRel, Oid classId, Oid objectId,
+static void shdepDropDependency(Relation sdepRel,
+					Oid classId, Oid objectId, int32 objsubId,
+					bool drop_subobjects,
 					Oid refclassId, Oid refobjId,
 					SharedDependencyType deptype);
 static void storeObjectDescription(StringInfo descs, objectType type,
@@ -111,6 +115,7 @@ recordSharedDependencyOn(ObjectAddress *depender,
 							  sdepRel))
 	{
 		shdepAddDependency(sdepRel, depender->classId, depender->objectId,
+						   depender->objectSubId,
 						   referenced->classId, referenced->objectId,
 						   deptype);
 	}
@@ -163,14 +168,15 @@ recordDependencyOnOwner(Oid classId, Oid objectId, Oid owner)
  * locked.
  */
 static void
-shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
+shdepChangeDep(Relation sdepRel,
+			   Oid classid, Oid objid, int32 objsubid,
 			   Oid refclassid, Oid refobjid,
 			   SharedDependencyType deptype)
 {
 	Oid			dbid = classIdGetDbId(classid);
 	HeapTuple	oldtup = NULL;
 	HeapTuple	scantup;
-	ScanKeyData key[3];
+	ScanKeyData key[4];
 	SysScanDesc scan;
 
 	/*
@@ -194,9 +200,13 @@ shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
 				Anum_pg_shdepend_objid,
 				BTEqualStrategyNumber, F_OIDEQ,
 				ObjectIdGetDatum(objid));
+	ScanKeyInit(&key[3],
+				Anum_pg_shdepend_objsubid,
+				BTEqualStrategyNumber, F_INT4EQ,
+				Int32GetDatum(objsubid));
 
 	scan = systable_beginscan(sdepRel, SharedDependDependerIndexId, true,
-							  SnapshotNow, 3, key);
+							  SnapshotNow, 4, key);
 
 	while ((scantup = systable_getnext(scan)) != NULL)
 	{
@@ -206,8 +216,8 @@ shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
 		/* Caller screwed up if multiple matches */
 		if (oldtup)
 			elog(ERROR,
-				 "multiple pg_shdepend entries for object %u/%u deptype %c",
-				 classid, objid, deptype);
+			   "multiple pg_shdepend entries for object %u/%u/%d deptype %c",
+				 classid, objid, objsubid, deptype);
 		oldtup = heap_copytuple(scantup);
 	}
 
@@ -239,11 +249,12 @@ shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
 		Datum		values[Natts_pg_shdepend];
 		bool		nulls[Natts_pg_shdepend];
 
-		memset(nulls, 0, sizeof(nulls));
+		memset(nulls, false, sizeof(nulls));
 
 		values[Anum_pg_shdepend_dbid - 1] = ObjectIdGetDatum(dbid);
 		values[Anum_pg_shdepend_classid - 1] = ObjectIdGetDatum(classid);
 		values[Anum_pg_shdepend_objid - 1] = ObjectIdGetDatum(objid);
+		values[Anum_pg_shdepend_objsubid - 1] = Int32GetDatum(objsubid);
 
 		values[Anum_pg_shdepend_refclassid - 1] = ObjectIdGetDatum(refclassid);
 		values[Anum_pg_shdepend_refobjid - 1] = ObjectIdGetDatum(refobjid);
@@ -268,6 +279,9 @@ shdepChangeDep(Relation sdepRel, Oid classid, Oid objid,
  * changeDependencyOnOwner
  *
  * Update the shared dependencies to account for the new owner.
+ *
+ * Note: we don't need an objsubid argument because only whole objects
+ * have owners.
  */
 void
 changeDependencyOnOwner(Oid classId, Oid objectId, Oid newOwnerId)
@@ -277,7 +291,8 @@ changeDependencyOnOwner(Oid classId, Oid objectId, Oid newOwnerId)
 	sdepRel = heap_open(SharedDependRelationId, RowExclusiveLock);
 
 	/* Adjust the SHARED_DEPENDENCY_OWNER entry */
-	shdepChangeDep(sdepRel, classId, objectId,
+	shdepChangeDep(sdepRel,
+				   classId, objectId, 0,
 				   AuthIdRelationId, newOwnerId,
 				   SHARED_DEPENDENCY_OWNER);
 
@@ -300,7 +315,7 @@ changeDependencyOnOwner(Oid classId, Oid objectId, Oid newOwnerId)
 	 * to make the various ALTER OWNER routines each know about it.
 	 *----------
 	 */
-	shdepDropDependency(sdepRel, classId, objectId,
+	shdepDropDependency(sdepRel, classId, objectId, 0, true,
 						AuthIdRelationId, newOwnerId,
 						SHARED_DEPENDENCY_ACL);
 
@@ -368,7 +383,7 @@ getOidListDiff(Oid *list1, int nlist1, Oid *list2, int nlist2, Oid **diff)
  * updateAclDependencies
  *		Update the pg_shdepend info for an object's ACL during GRANT/REVOKE.
  *
- * classId, objectId: identify the object whose ACL this is
+ * classId, objectId, objsubId: identify the object whose ACL this is
  * ownerId: role owning the object
  * isGrant: are we adding or removing ACL entries?
  * noldmembers, oldmembers: array of roleids appearing in old ACL
@@ -388,7 +403,8 @@ getOidListDiff(Oid *list1, int nlist1, Oid *list2, int nlist2, Oid **diff)
  * before return.
  */
 void
-updateAclDependencies(Oid classId, Oid objectId, Oid ownerId, bool isGrant,
+updateAclDependencies(Oid classId, Oid objectId, int32 objsubId,
+					  Oid ownerId, bool isGrant,
 					  int noldmembers, Oid *oldmembers,
 					  int nnewmembers, Oid *newmembers)
 {
@@ -429,11 +445,12 @@ updateAclDependencies(Oid classId, Oid objectId, Oid ownerId, bool isGrant,
 				continue;
 
 			if (isGrant)
-				shdepAddDependency(sdepRel, classId, objectId,
+				shdepAddDependency(sdepRel, classId, objectId, objsubId,
 								   AuthIdRelationId, roleid,
 								   SHARED_DEPENDENCY_ACL);
 			else
-				shdepDropDependency(sdepRel, classId, objectId,
+				shdepDropDependency(sdepRel, classId, objectId, objsubId,
+									false,		/* exact match on objsubId */
 									AuthIdRelationId, roleid,
 									SHARED_DEPENDENCY_ACL);
 		}
@@ -457,11 +474,14 @@ typedef struct
  * checkSharedDependencies
  *
  * Check whether there are shared dependency entries for a given shared
- * object.	Returns a string containing a newline-separated list of object
+ * object; return true if so.
+ *
+ * In addition, return a string containing a newline-separated list of object
  * descriptions that depend on the shared object, or NULL if none is found.
- * The size of the returned string is limited to about MAX_REPORTED_DEPS lines;
- * if there are more objects than that, the output is returned truncated at
- * that point while the full message is logged to the postmaster log.
+ * We actually return two such strings; the "detail" result is suitable for
+ * returning to the client as an errdetail() string, and is limited in size.
+ * The "detail_log" string is potentially much longer, and should be emitted
+ * to the server log only.
  *
  * We can find three different kinds of dependencies: dependencies on objects
  * of the current database; dependencies on shared objects; and dependencies
@@ -471,8 +491,9 @@ typedef struct
  *
  * If we find a SHARED_DEPENDENCY_PIN entry, we can error out early.
  */
-char *
-checkSharedDependencies(Oid classId, Oid objectId)
+bool
+checkSharedDependencies(Oid classId, Oid objectId,
+						char **detail_msg, char **detail_log_msg)
 {
 	Relation	sdepRel;
 	ScanKeyData key[2];
@@ -490,9 +511,7 @@ checkSharedDependencies(Oid classId, Oid objectId)
 	/*
 	 * We limit the number of dependencies reported to the client to
 	 * MAX_REPORTED_DEPS, since client software may not deal well with
-	 * enormous error strings.	The server log always gets a full report,
-	 * which is collected in a separate StringInfo if and only if we detect
-	 * that the client report is going to be truncated.
+	 * enormous error strings.	The server log always gets a full report.
 	 */
 #define MAX_REPORTED_DEPS 100
 
@@ -531,7 +550,7 @@ checkSharedDependencies(Oid classId, Oid objectId)
 
 		object.classId = sdepForm->classid;
 		object.objectId = sdepForm->objid;
-		object.objectSubId = 0;
+		object.objectSubId = sdepForm->objsubid;
 
 		/*
 		 * If it's a dependency local to this database or it's a shared
@@ -549,15 +568,9 @@ checkSharedDependencies(Oid classId, Oid objectId)
 									   sdepForm->deptype, 0);
 			}
 			else
-			{
 				numNotReportedDeps++;
-				/* initialize the server-only log line */
-				if (alldescs.len == 0)
-					appendBinaryStringInfo(&alldescs, descs.data, descs.len);
-
-				storeObjectDescription(&alldescs, LOCAL_OBJECT, &object,
-									   sdepForm->deptype, 0);
-			}
+			storeObjectDescription(&alldescs, LOCAL_OBJECT, &object,
+								   sdepForm->deptype, 0);
 		}
 		else if (sdepForm->dbid == InvalidOid)
 		{
@@ -568,15 +581,9 @@ checkSharedDependencies(Oid classId, Oid objectId)
 									   sdepForm->deptype, 0);
 			}
 			else
-			{
 				numNotReportedDeps++;
-				/* initialize the server-only log line */
-				if (alldescs.len == 0)
-					appendBinaryStringInfo(&alldescs, descs.data, descs.len);
-
-				storeObjectDescription(&alldescs, SHARED_OBJECT, &object,
-									   sdepForm->deptype, 0);
-			}
+			storeObjectDescription(&alldescs, SHARED_OBJECT, &object,
+								   sdepForm->deptype, 0);
 		}
 		else
 		{
@@ -615,9 +622,7 @@ checkSharedDependencies(Oid classId, Oid objectId)
 	heap_close(sdepRel, AccessShareLock);
 
 	/*
-	 * Report dependencies on remote databases.  If we're truncating the
-	 * output already, don't put a line per database, but a single one for all
-	 * of them.  Otherwise add as many as fit in MAX_REPORTED_DEPS.
+	 * Summarize dependencies in remote databases.
 	 */
 	foreach(cell, remDeps)
 	{
@@ -634,15 +639,9 @@ checkSharedDependencies(Oid classId, Oid objectId)
 								   SHARED_DEPENDENCY_INVALID, dep->count);
 		}
 		else
-		{
 			numNotReportedDbs++;
-			/* initialize the server-only log line */
-			if (alldescs.len == 0)
-				appendBinaryStringInfo(&alldescs, descs.data, descs.len);
-
-			storeObjectDescription(&alldescs, REMOTE_OBJECT, &object,
-								   SHARED_DEPENDENCY_INVALID, dep->count);
-		}
+		storeObjectDescription(&alldescs, REMOTE_OBJECT, &object,
+							   SHARED_DEPENDENCY_INVALID, dep->count);
 	}
 
 	list_free_deep(remDeps);
@@ -651,34 +650,28 @@ checkSharedDependencies(Oid classId, Oid objectId)
 	{
 		pfree(descs.data);
 		pfree(alldescs.data);
-		return NULL;
+		*detail_msg = *detail_log_msg = NULL;
+		return false;
 	}
 
 	if (numNotReportedDeps > 0)
-		appendStringInfo(&descs, _("\nand %d other objects "
-								   "(see server log for list)"),
+		appendStringInfo(&descs, ngettext("\nand %d other object "
+										  "(see server log for list)",
+										  "\nand %d other objects "
+										  "(see server log for list)",
+										  numNotReportedDeps),
 						 numNotReportedDeps);
 	if (numNotReportedDbs > 0)
-		appendStringInfo(&descs, _("\nand objects in %d other databases "
-								   "(see server log for list)"),
+		appendStringInfo(&descs, ngettext("\nand objects in %d other database "
+										  "(see server log for list)",
+									   "\nand objects in %d other databases "
+										  "(see server log for list)",
+										  numNotReportedDbs),
 						 numNotReportedDbs);
 
-	if (numNotReportedDeps > 0 || numNotReportedDbs > 0)
-	{
-		ObjectAddress obj;
-
-		obj.classId = classId;
-		obj.objectId = objectId;
-		obj.objectSubId = 0;
-		ereport(LOG,
-				(errmsg("there are objects dependent on %s",
-						getObjectDescription(&obj)),
-				 errdetail("%s", alldescs.data)));
-	}
-
-	pfree(alldescs.data);
-
-	return descs.data;
+	*detail_msg = descs.data;
+	*detail_log_msg = alldescs.data;
+	return true;
 }
 
 /*
@@ -785,7 +778,7 @@ dropDatabaseDependencies(Oid databaseId)
 	systable_endscan(scan);
 
 	/* Now delete all entries corresponding to the database itself */
-	shdepDropDependency(sdepRel, DatabaseRelationId, databaseId,
+	shdepDropDependency(sdepRel, DatabaseRelationId, databaseId, 0, true,
 						InvalidOid, InvalidOid,
 						SHARED_DEPENDENCY_INVALID);
 
@@ -798,15 +791,19 @@ dropDatabaseDependencies(Oid databaseId)
  * Delete all pg_shdepend entries corresponding to an object that's being
  * dropped or modified.  The object is assumed to be either a shared object
  * or local to the current database (the classId tells us which).
+ *
+ * If objectSubId is zero, we are deleting a whole object, so get rid of
+ * pg_shdepend entries for subobjects as well.
  */
 void
-deleteSharedDependencyRecordsFor(Oid classId, Oid objectId)
+deleteSharedDependencyRecordsFor(Oid classId, Oid objectId, int32 objectSubId)
 {
 	Relation	sdepRel;
 
 	sdepRel = heap_open(SharedDependRelationId, RowExclusiveLock);
 
-	shdepDropDependency(sdepRel, classId, objectId,
+	shdepDropDependency(sdepRel, classId, objectId, objectSubId,
+						(objectSubId == 0),
 						InvalidOid, InvalidOid,
 						SHARED_DEPENDENCY_INVALID);
 
@@ -821,7 +818,8 @@ deleteSharedDependencyRecordsFor(Oid classId, Oid objectId)
  * locked.
  */
 static void
-shdepAddDependency(Relation sdepRel, Oid classId, Oid objectId,
+shdepAddDependency(Relation sdepRel,
+				   Oid classId, Oid objectId, int32 objsubId,
 				   Oid refclassId, Oid refobjId,
 				   SharedDependencyType deptype)
 {
@@ -844,6 +842,7 @@ shdepAddDependency(Relation sdepRel, Oid classId, Oid objectId,
 	values[Anum_pg_shdepend_dbid - 1] = ObjectIdGetDatum(classIdGetDbId(classId));
 	values[Anum_pg_shdepend_classid - 1] = ObjectIdGetDatum(classId);
 	values[Anum_pg_shdepend_objid - 1] = ObjectIdGetDatum(objectId);
+	values[Anum_pg_shdepend_objsubid - 1] = Int32GetDatum(objsubId);
 
 	values[Anum_pg_shdepend_refclassid - 1] = ObjectIdGetDatum(refclassId);
 	values[Anum_pg_shdepend_refobjid - 1] = ObjectIdGetDatum(refobjId);
@@ -865,20 +864,26 @@ shdepAddDependency(Relation sdepRel, Oid classId, Oid objectId,
  *		Internal workhorse for deleting entries from pg_shdepend.
  *
  * We drop entries having the following properties:
- *	dependent object is the one identified by classId/objectId
+ *	dependent object is the one identified by classId/objectId/objsubId
  *	if refclassId isn't InvalidOid, it must match the entry's refclassid
  *	if refobjId isn't InvalidOid, it must match the entry's refobjid
  *	if deptype isn't SHARED_DEPENDENCY_INVALID, it must match entry's deptype
  *
+ * If drop_subobjects is true, we ignore objsubId and consider all entries
+ * matching classId/objectId.
+ *
  * sdepRel must be the pg_shdepend relation, already opened and suitably
  * locked.
  */
 static void
-shdepDropDependency(Relation sdepRel, Oid classId, Oid objectId,
+shdepDropDependency(Relation sdepRel,
+					Oid classId, Oid objectId, int32 objsubId,
+					bool drop_subobjects,
 					Oid refclassId, Oid refobjId,
 					SharedDependencyType deptype)
 {
-	ScanKeyData key[3];
+	ScanKeyData key[4];
+	int			nkeys;
 	SysScanDesc scan;
 	HeapTuple	tup;
 
@@ -895,9 +900,19 @@ shdepDropDependency(Relation sdepRel, Oid classId, Oid objectId,
 				Anum_pg_shdepend_objid,
 				BTEqualStrategyNumber, F_OIDEQ,
 				ObjectIdGetDatum(objectId));
+	if (drop_subobjects)
+		nkeys = 3;
+	else
+	{
+		ScanKeyInit(&key[3],
+					Anum_pg_shdepend_objsubid,
+					BTEqualStrategyNumber, F_INT4EQ,
+					Int32GetDatum(objsubId));
+		nkeys = 4;
+	}
 
 	scan = systable_beginscan(sdepRel, SharedDependDependerIndexId, true,
-							  SnapshotNow, 3, key);
+							  SnapshotNow, nkeys, key);
 
 	while (HeapTupleIsValid(tup = systable_getnext(scan)))
 	{
@@ -1034,7 +1049,10 @@ storeObjectDescription(StringInfo descs, objectType type,
 
 		case REMOTE_OBJECT:
 			/* translator: %s will always be "database %s" */
-			appendStringInfo(descs, _("%d objects in %s"), count, objdesc);
+			appendStringInfo(descs, ngettext("%d object in %s",
+											 "%d objects in %s",
+											 count),
+							 count, objdesc);
 			break;
 
 		default:
@@ -1097,7 +1115,7 @@ isSharedObjectPinned(Oid classId, Oid objectId, Relation sdepRel)
  *
  * Drop the objects owned by any one of the given RoleIds.	If a role has
  * access to an object, the grant will be removed as well (but the object
- * will not, of course.)
+ * will not, of course).
  *
  * We can revoke grants immediately while doing the scan, but drops are
  * saved up and done all at once with performMultipleDeletions.  This
@@ -1161,17 +1179,12 @@ shdepDropOwned(List *roleids, DropBehavior behavior)
 
 		while ((tuple = systable_getnext(scan)) != NULL)
 		{
-			ObjectAddress obj;
-			GrantObjectType objtype;
-			InternalGrant istmt;
 			Form_pg_shdepend sdepForm = (Form_pg_shdepend) GETSTRUCT(tuple);
+			InternalGrant istmt;
+			ObjectAddress obj;
 
-			/*
-			 * We only operate on shared objects and objects in the current
-			 * database
-			 */
-			if (sdepForm->dbid != MyDatabaseId &&
-				sdepForm->dbid != InvalidOid)
+			/* We only operate on objects in the current database */
+			if (sdepForm->dbid != MyDatabaseId)
 				continue;
 
 			switch (sdepForm->deptype)
@@ -1206,14 +1219,13 @@ shdepDropOwned(List *roleids, DropBehavior behavior)
 						default:
 							elog(ERROR, "unexpected object type %d",
 								 sdepForm->classid);
-							/* keep compiler quiet */
-							objtype = (GrantObjectType) 0;
 							break;
 					}
 					istmt.is_grant = false;
 					istmt.objects = list_make1_oid(sdepForm->objid);
 					istmt.all_privs = true;
 					istmt.privileges = ACL_NO_RIGHTS;
+					istmt.col_privs = NIL;
 					istmt.grantees = list_make1_oid(roleid);
 					istmt.grant_option = false;
 					istmt.behavior = DROP_CASCADE;
@@ -1224,7 +1236,7 @@ shdepDropOwned(List *roleids, DropBehavior behavior)
 					/* Save it for deletion below */
 					obj.classId = sdepForm->classid;
 					obj.objectId = sdepForm->objid;
-					obj.objectSubId = 0;
+					obj.objectSubId = sdepForm->objsubid;
 					add_exact_object_address(&obj, deleteobjs);
 					break;
 			}
@@ -1278,7 +1290,8 @@ shdepReassignOwned(List *roleids, Oid newrole)
 
 			ereport(ERROR,
 					(errcode(ERRCODE_DEPENDENT_OBJECTS_STILL_EXIST),
-					 errmsg("cannot reassign ownership of objects owned by %s because they are required by the database system",
+				   errmsg("cannot drop objects owned by %s because they are "
+						  "required by the database system",
 						  getObjectDescription(&obj))));
 
 			/*
@@ -1352,14 +1365,6 @@ shdepReassignOwned(List *roleids, Oid newrole)
 					AlterLanguageOwner_oid(sdepForm->objid, newrole);
 					break;
 
-				case OperatorClassRelationId:
-					AlterOpClassOwner_oid(sdepForm->objid, newrole);
-					break;
-
-				case OperatorFamilyRelationId:
-					AlterOpFamilyOwner_oid(sdepForm->objid, newrole);
-					break;
-
 				default:
 					elog(ERROR, "unexpected classid %d", sdepForm->classid);
 					break;
