@@ -2,9 +2,9 @@
  *
  * PostgreSQL locale utilities
  *
- * $Id: pg_locale.h,v 1.15.4.1 2006/01/05 00:55:24 tgl Exp $
+ * $PostgreSQL: pgsql/src/include/utils/pg_locale.h,v 1.27 2009/01/01 17:24:02 momjian Exp $
  *
- * Copyright (c) 2002-2003, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2009, PostgreSQL Global Development Group
  *
  *-----------------------------------------------------------------------
  */
@@ -14,23 +14,36 @@
 
 #include <locale.h>
 
+#include "utils/guc.h"
+
+
+/* GUC settings */
 extern char *locale_messages;
 extern char *locale_monetary;
 extern char *locale_numeric;
 extern char *locale_time;
 
-extern const char *locale_messages_assign(const char *value,
-					   bool doit, bool interactive);
-extern const char *locale_monetary_assign(const char *value,
-					   bool doit, bool interactive);
-extern const char *locale_numeric_assign(const char *value,
-					  bool doit, bool interactive);
-extern const char *locale_time_assign(const char *value,
-				   bool doit, bool interactive);
+/* lc_time localization cache */
+extern char *localized_abbrev_days[];
+extern char *localized_full_days[];
+extern char *localized_abbrev_months[];
+extern char *localized_full_months[];
 
+
+extern const char *locale_messages_assign(const char *value,
+					   bool doit, GucSource source);
+extern const char *locale_monetary_assign(const char *value,
+					   bool doit, GucSource source);
+extern const char *locale_numeric_assign(const char *value,
+					  bool doit, GucSource source);
+extern const char *locale_time_assign(const char *value,
+				   bool doit, GucSource source);
+
+extern bool check_locale(int category, const char *locale);
 extern char *pg_perm_setlocale(int category, const char *locale);
 
 extern bool lc_collate_is_c(void);
+extern bool lc_ctype_is_c(void);
 
 /*
  * Return the POSIX lconv struct (contains number/money formatting
@@ -38,4 +51,6 @@ extern bool lc_collate_is_c(void);
  */
 extern struct lconv *PGLC_localeconv(void);
 
+extern void cache_locale_time(void);
+
 #endif   /* _PG_LOCALE_ */
